import hell;
import std.random;
import std.string;
import std.math;
import std.math2;

// 画面サイズ
const int g_width = 320;
const int g_height = 480;

int level = 0;
/**
 * ２次元ベクトル
 */
class Vec2D
{
public:
	/**
	 * 値
	 */
	float x, y;
	/**
	 * コンストラクタ
	 */
	this(float x=0, float y=0) { this.x = x; this.y = y; }
	void opAddAssign(Vec2D v)  { x += v.x; y += v.y; }
	void opSubAssign(Vec2D v)  { x -= v.x; y -= v.y; }
	void opMulAssign(float a)  { x *= a;   y *= a;   }
	void opDivAssign(float a)  { x /= a;   y /= a;   }
	/**
	 * 正規化
	 */
	void normalize()
	{
		float len = length();
		if(len == 0) return;
		x /= len; y /= len;
	}
	/**
	 * 距離
	 */
	float length()   { return sqrt(lengthSq());}
	/**
	 * 距離の２乗
	 */
	float lengthSq() { return x*x + y*y; }
}

/**
 * トークンクラス
 */
class Token
{
public:
	/**
	 * 座標
	 */
	float x, y;
	/**
	 * 移動量
	 */
	float vx, vy;
	/**
	 * 加速度
	 */
	float ax, ay;
	/**
	 * 生存フラグ
	 */
	bool exist;
	/**
	 * 当たり判定サイズ
	 */
	float w, h;
	/**
	 * 見た目サイズ
	 */
	float gw, gh;
public:
	/**
	 * コンストラクタ
	 */
	this()
	{
		exist = false;
	}
	/**
	 * 生成
	 */
	void create(float x, float y, float w, float h, float gw, float gh , float vx=0, float vy=0, float ax=0, float ay=0)
	{
		this.x = x; this.y = y;
		this.w = w; this.h = h;
		this.gw = gw; this.gh = gh;
		this.vx = vx; this.vy = vy;
		this.ax = ax; this.ay = ay;
		exist = true;
	}
	/**
	 * Xの中心点を取得
	 */
	float getX()
	{
		return x;
	}
	/**
	 * Yの中心点を取得
	 */
	float getY()
	{
		return y;
	}
	/**
	 * 移動
	 */
	void move()
	{
		vx += ax; vy += ay;
		x += vx;  y += vy;
	}
	/**
	 * 点との当たり判定
	 */
	bool isHit(int px, int py)
	{
		return x <= px && px <= x + w && y <= py && py <= y + h;
	}
	/**
	 * 別のトークン（矩形）との当たり判定
	 */
	bool isHit(Token t)
	{
		float left1   = x - w/2;
		float top1    = y - h/2;
		float right1  = x + w/2;
		float bottom1 = y + h/2;
		float left2   = t.x - t.w/2;
		float top2    = t.y - t.h/2;
		float right2  = t.x + t.w/2;
		float bottom2 = t.y + t.h/2;
		return (left1 < right2) && (right1 > left2) && (top1 < bottom2) && (bottom1 > top2);
	}
	/**
	 * ウィンドウサイズ内に存在するかどうか
	 */
	bool isOutOfWindow()
	{
		return x < -w/2 || x > g_width + w/2 || y < -h/2 || y > g_height + h/2;
	}
	void dump()
	{
		Hell_drawFont(
			std.string.toString(x) ~ "," ~ std.string.toString(y) ~ "," ~ std.string.toString(vx) ~ "," ~ std.string.toString(vy),
			0, 0
		);
	}
	/**
	 * 描画
	 */
	void draw(int r, int g, int b)
	{
		Hell_fillRect(cast(int)(x - gw/2) , cast(int)(y - gh/2) , cast(int)gw, cast(int)gh, r, g, b);
		//Hell_fillRect(cast(int)(x -  w/2) , cast(int)(y -  h/2) , cast(int)w, cast(int)h, 255, g, b);		// 当たり判定確認
	}
}

/**
 * トークンの管理クラス
 */
class TokenManager(T)
{
public:
	/**
	 * 保持しているインスタンス
	 */
	T[] _pool;
public:
	/**
	 * コンストラクタ
	 */
	this(int size)
	{
		_pool = new T[size];
		foreach(inout t; _pool)
		{
			t = new T;
		}
	}
	/**
	 * 利用可能なトークンの取得
	 */
	T getToken()
	{
		foreach(t; _pool)
		{
			if(!t.exist) return t;
		}
		return null;
	}
	/**
	 * 保持しているトークン配列の取得
	 */
	T[] getArray() { return _pool; }
	/**
	 * 更新
	 */
	void update()
	{
		foreach(t; _pool)
		{
			if(!t.exist) continue;
			t.update();
		}
	}
	/**
	 * 描画
	 */
	void draw()
	{
		foreach(t; _pool)
		{
			if(!t.exist) continue;
			t.draw();
		}
	}
	/**
	 * 全てを消滅
	 */
	void vanish()
	{
		foreach(t; _pool)
		{
			t.exist = false;
		}
	}
	
	/**
	 * 生存数
	 */
	int getLives()
	{
		int result = 0;
		foreach(t; _pool)
		{
			if(t.exist) result++;
		}
		return result;
	}
}

/**
 * 敵弾クラス
 */
class Bullet : Token
{
public:
	/**
	 * 親トークン管理
	 */
	static TokenManager!(Bullet) parent;
	/**
	 * プレイヤー
	 */
	static Player target;
	int id;     // ID
	int timer;
public:
	/**
	 * 狙い撃ち角度
	 */
	float getAim()
	{
		float dx = target.x - x;
		float dy = target.y - y;
		return rad2deg(atan2(-dy, dx));
	}
	/**
	 * 生成
	 */
	void create(int id, float x, float y, float w, float h, float deg, float speed, float ax=0, float ay=0)
	{
		this.id = id;
		timer = 0;
		super.create(x, y, w, h, w , h , cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed, ax, ay);
	}
	/**
	 * 生成（外から呼ばれる）
	 */
	static void add(float x, float y, float deg, float speed, float ax=0, float ay=0)
	{
		if(level % 5 == 2 && level < 20)speed *= 1.5;		//弾速up
		if(level % 5 == 2 && level > 19)speed *= 1.8;		//弾速up
		
		Bullet bullet = parent.getToken();
		if(!bullet) return;
		bullet.create(0, x, y, 8, 8, deg, speed, ax, ay);
		
		if(level % 5 == 3){
			bullet = parent.getToken();
			if(!bullet) return;
			bullet.create(0, x, y, 8, 8, deg, speed * 0.8, ax, ay);
		};
	}
	static void addNormal(float x, float y, float deg, float speed, float ax=0, float ay=0)
	{
		Bullet bullet = parent.getToken();
		if(!bullet) return;
		bullet.create(0, x, y, 8, 8, deg, speed, ax, ay);
	}
	/**
	 * 更新
	 */
	void update()
	{
		super.move();
		if(isOutOfWindow())
		{
			exist = false;
		}
		timer++;
	}
	/**
	 * 描画
	 */
	void draw()
	{
//		dump();
//		super.draw(0xff, 0, 0);

		int t = timer / 3;
		if(t%4 == 0) Hell_fillRect(cast(int)(x - gw/2 - 1) , cast(int)(y - gh/2 + 1) , cast(int)gw + 2, cast(int)gh - 2, 0xff,0x80,0x00);
		if(t%4 == 1 | t%4 == 3) super.draw(0xff,0x40,0x00);
		if(t%4 == 2) Hell_fillRect(cast(int)(x - gw/2 + 1) , cast(int)(y - gh/2 - 1) , cast(int)gw - 2, cast(int)gh + 2, 0xff,0x00,0x00);
	}
}

/**
 * 自機ショットクラス
 */
class Shot : Token
{
public:
	/**
	 * 親
	 */
	static TokenManager!(Shot) parent;
public:
	/**
	 * 生成
	 */
	void create(float x, float y, float deg, float speed)
	{
		super.create(x, y, 4, 4, 4, 4, cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed);
	}
	/**
	 * 生成（外からの入り口）
	 */
	static void add(float x, float y, float deg, float speed)
	{
		Shot shot = parent.getToken();
		if(!shot) return;
		shot.create(x, y, deg, speed);
	}
	/**
	 * 更新
	 */
	void update()
	{
		super.move();
		if(isOutOfWindow())
		{
			exist = false;
		}
	}
	/**
	 * 描画
	 */
	void draw()
	{
		super.draw(0x80, 0x80, 0xff);
	}
}

/**
 * 自機クラス
 */
class Player : Token
{
public:
	/**
	 * ショット連射性能（間隔）初期値
	 */
	static const int SHOT_COUNTER_INIT = 5;
	/**
	 * 打ち返しに必要な時間（240/120fps=2sec）
	 */
	static const int BOMB_COUNTER_WAIT = 120;	// 240;
	/**
	 * 打ち返し最大使用回数
	 */
	static const int BOMB_USE_COUNT = 2;
	/**
	 * 打ち返しショットのサイズ
	 */
	static const int BOMB_SIZE = 48;
	/**
	 * ノーマル時の移動の速さ
	 */
	static const float SPEED_NORMAL = 5;
	/**
	 * ショット時の移動の速さ
	 */
	static const float SPEED_SLOW = 2;
	/**
	 * ワインダー角度
	 */
	float armDeg;
	/**
	 * アーム左
	 */
	Token armL;
	/**
	 * アーム右
	 */
	Token armR;
	/**
	 * ショット間隔タイマー
	 */
	int tShot;
	/**
	 * ショット間隔タイマー2
	 */
	int tShot2;
	/**
	 * ショット可能時間
	 */
	int tShotCounter;
	/**
	 * 打ち返しタイマー
	 */
	int tBomb;
	/**
	 * アーム存在フラグ
	 */
	bool arm;
	/**
	 * 打ち返し当たり判定用トークン（左）
	 */
	Token bombL;
	/**
	 * 打ち返し当たり判定用トークン（右）
	 */
	Token bombR;
public:
	/**
	 * コンストラクタ
	 */
	this()
	{
		armL = new Token();
		armR = new Token();
		bombL = new Token();
		bombR = new Token();
	}
	/**
	 * 生成
	 */
	void create(float x, float y, float w, float h)
	{
		tBomb = BOMB_COUNTER_WAIT;
		tShotCounter = SHOT_COUNTER_INIT;	//10;
		armDeg = 90;
		arm = true;
		super.create(x, y, 2, 2, w, h);
		armL.create(x, y, 8, 8 , 8 , 8);
		armR.create(x, y, 8, 8 , 8 , 8);
	}
	/**
	 * ダメージ
	 */
	void damage()
	{
		Hell_fillRect(0, 0, 0, 0, 0xff, 0xff, 0xff);
		g_gamemain.bonus_ratio /= 2;
		if(g_gamemain.item_gain > 0) g_gamemain.item_gain -= 1;
		if(arm)
		{
			// パーツ破損
			tBomb = 0;
			arm = false;
			tShotCounter = SHOT_COUNTER_INIT;
			Particle.addDamage(armR.getX(), armR.getY(), 4);
			Particle.addDamage(armL.getX(), armL.getY(), 4);
		}
		else
		{
			// 死亡
			Particle.addDamage(getX(), getY(), 8);
			exist = false;
		}
		Hell_playWAV("damage");
	}
	/**
	 * アイテム取得
	 */
	void powerup()
	{
		Hell_playWAV("kin");
		tBomb += BOMB_COUNTER_WAIT;
		if(tBomb > BOMB_COUNTER_WAIT*BOMB_USE_COUNT) tBomb = BOMB_COUNTER_WAIT*BOMB_USE_COUNT;
		if(!arm)
		{
			Particle.addRecovery(getX(), getY());
			arm = true;
			return;
		}
		// ショットパワーアップ
		//tShotCounter -= 2;
		//if(tShotCounter < 3)
		//{
			// ボーナス
			//tShotCounter = 3;
			Particle.addMessage(getX(), getY(), "500");
			g_gamemain.addScore(500);
			g_gamemain.item_gain++;
		//}
		//else
		//{
		//	Particle.addItem(getX(), getY());
		//}
	}
	/**
	 * 更新
	 */
	void update()
	{
		if(g_gamemain.bonus_ratio > 10) g_gamemain.bonus_ratio--;
		if(g_gamemain.bonus_ratio < 10) g_gamemain.bonus_ratio++;
		
		if(tShot < 1)
		{
			if(isPressEnter())
			{
				if(arm)
				{
				//	Shot.add(armL.getX(), armL.getY(), 180-armDeg, 10);
				//	Shot.add(armR.getX(), armR.getY(), armDeg, 10);
				}
				else
				{
					Shot.add(getX() - 4, getY(), 90, 10);Shot.add(getX() + 4, getY(), 90, 10);
					tShot2++;
					Shot.add(getX(), getY(), 90-5*(tShot2 % 8), 10);Shot.add(getX(), getY(), 90+5*(tShot2 % 8), 10);
				}
				tShot = 2;//tShotCounter;
			}
		}
		else
		{
			tShot--;
		}
		if(arm)
		{
			if(tBomb >= BOMB_COUNTER_WAIT)
			{
				tBomb++;
				if(tBomb > BOMB_COUNTER_WAIT*BOMB_USE_COUNT) tBomb = BOMB_COUNTER_WAIT*BOMB_USE_COUNT;
				if(isPushEnter())
				{
					Hell_playWAV("winder");
					tBomb -= BOMB_COUNTER_WAIT;
					Particle.addWinder(armL.getX(), armL.getY());
					Particle.addWinder(armR.getX(), armR.getY());
					bombL.create(armL.getX(), armL.getY(), BOMB_SIZE, BOMB_SIZE, BOMB_SIZE, BOMB_SIZE);
					bombR.create(armR.getX(), armR.getY(), BOMB_SIZE, BOMB_SIZE, BOMB_SIZE, BOMB_SIZE);
					
					//Bomb.add(x, y, 90, 10);
					
					foreach(bullet; Bullet.parent.getArray())
					{
						if(!bullet.exist) continue;
						bool hitL = bombL.isHit(bullet);
						bool hitR = bombR.isHit(bullet);
						if(hitL && hitR)
						{
							Vec2D vL = new Vec2D(bombL.getX() - bullet.getX(), bombL.getY() - bullet.getY());
							Vec2D vR = new Vec2D(bombR.getX() - bullet.getX(), bombR.getY() - bullet.getY());
							if(vL.length() > vR.length()) hitL = false;
							else                          hitR = false;
						}
						if(hitL)
						{
							Bomb.add(bullet.x, bullet.y, 180-armDeg, 10);
							g_gamemain.counter++;
							g_gamemain.bonus_ratio += 20;
							bullet.exist = false;
						}
						if(hitR)
						{
							Bomb.add(bullet.x, bullet.y, armDeg, 10);
							g_gamemain.counter++;
							g_gamemain.bonus_ratio += 20;
							bullet.exist = false;
						}
					}
					if(g_gamemain.bonus_ratio > 199) g_gamemain.bonus_ratio = 199;
				}
			}
			else
			{
				tBomb++;
			}
		}
		move();
	}
	override void move()
	{
		// 移動
		float MOVE = SPEED_NORMAL;
		if(isPressEnter()) MOVE = SPEED_SLOW;
		Vec2D d = new Vec2D();
		if(isPressLeft())  d.x = -1;
		if(isPressUp())    d.y = -1;
		if(isPressRight()) d.x = 1;
		if(isPressDown())  d.y = 1;
		d.normalize();
		x += d.x*MOVE;
		y += d.y*MOVE;
		// 画面内に戻す
		if(x < 0) x = 0;
		if(x > g_width-w) x = g_width-w;
		if(y < 0) y = 0;
		if(y > g_height-h) y = g_height-h;
		if(arm) updateArm();
	}
	
	/**
	 * アーム更新
	 */
	void updateArm()
	{
/*
		if(isPressEnter())
		{
			const float MOVE = 3;
			if(isPressUp())   armDeg -= MOVE;
			if(isPressDown()) armDeg += MOVE;
			if(armDeg > 90) armDeg = 90;
			if(armDeg < 0)  armDeg = 0;
		}
*/		
		armL.x = getX() -5 - 32*cos(deg2rad(armDeg));
		armL.y = getY() - 4    - 32*sin(deg2rad(armDeg));
		armR.x = getX() +5 + 32*cos(deg2rad(armDeg));
		armR.y = armL.y;
	}
	/**
	 * 描画
	 */
	void draw()
	{
		super.draw(0x40, 0x40, 0xff);
		if(!arm) return;
		// アーム描画
		if(isPressEnter())
		{
			armL.draw(0x80, 0x80, 0xff);
			armR.draw(0x80, 0x80, 0xff);
		}
		else
		{
			armL.draw(0xff, 0xff, 0);
			armR.draw(0xff, 0xff, 0);
		}
		// 打ち返しメータ描画
		int[] c;
		if(tBomb >= BOMB_COUNTER_WAIT) c = [rand()%255, rand()%255, rand()%255];
		else                           c = [0, 0, 0x80];
		Hell_fillRect(10, 450, 300*tBomb/(BOMB_COUNTER_WAIT*BOMB_USE_COUNT), 5, c[0], c[1], c[2]);
	}
}

/**
 * 敵トークン
 */
class Enemy : Token
{
public:
	/**
	 * 親
	 */
	static TokenManager!(Enemy) parent;
	/**
	 * 自機
	 */
	static Player target;
	/**
	 * HPテーブル
	 */
	static int[] hpTbl = [
		// 0,    1,  2, 3,  4,  5,   6,  7,   8,  9,   10   11  , 12 , 13, 14 , 15
		// 1000, 20, 3, 50, 50, 120, 30, 200, 50, 200, 200, 
		   1600, 60, 3, 60, 60, 60, 30, 60, 60,60, 60, 60 , 60, 60 , 50 , 60 , 60
	];
	/**
	 * スコアテーブル
	 */
	static int[] scrTbl = [
		// 0,    1,  2,  3,  4,  5,   6,  7,   8,  9, 10   11 , 12 , 13  , 14 , 15
		   2000,100, 10,500,500,100, 10,100, 100,100, 100, 500, 100, 200 ,100 ,100 ,100
	];
	/**
	 * ID
	 */
	int id;
	/**
	 * 耐久力
	 */
	int hp;
	/**
	 * 耐久力初期値
	 */
	int hpmax;
	/**
	 * スコア）
	 */
	int score;
	/**
	 * タイマ
	 */
	int timer;
public:
	/**
	 * 狙い撃ち角度の取得
	 */
	float getAim()
	{
		float dx = target.x - x;
		float dy = target.y - y;
		return rad2deg(atan2(-dy, dx));
	}
	/**
	 * 生成
	 */
	void create(int id, float x, float y, float w, float h, float deg=0, float speed=0)
	{
		this.id = id;
		hpmax = hpTbl[id];
		hp = hpmax;
		super.create(x, y, w, h, w, h, cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed);
		
		timer = 0;
		if(level%2 == 0 && level > 20) timer = 10;
		switch(id)
		{
		case 0:
			//timer = 3500;
			hpmax = hpmax + level * 200;
			timer = 0;
			hp = hpmax;
			break;
		case 2,14:
			timer = rand()%30;
			break;
		default:
			break;
		}
	}
	/**
	 * 生成（外から呼ばれる）
	 */
	static void add(int id, float x, float y, float w, float h, float deg=0, float speed=0)
	{
		Enemy enemy = parent.getToken();
		if(!enemy) return;
		enemy.create(id, x, y, w, h, deg, speed);
	}
	/**
	 * ダメージ
	 */
	void damage(int param)
	{
		hp -= param;
		if(hp < 1)
		{
			hp = 0;
			vanish();
			
			int r = cast(int)(g_gamemain.getBonusRatio());
			if(r < 2){
				Particle.addMessage(getX(), y+h, std.string.toString(scrTbl[id]));
				g_gamemain.addScore(scrTbl[id]);
			}else{
				Particle.addMessage(getX(), y+h, std.string.toString(scrTbl[id]) ~ " x " ~ std.string.toString(r));
				g_gamemain.addScore(scrTbl[id] * r);
			}
			
			
			// 破壊打ち返し
			if(level > 14 & id != 2){
				if(level % 5 == 2){
					for(int i = 0; i < 18; i++){Bullet.addNormal(getX(), getY(), i * 20 ,2);}
				}else{
					for(int i = 0; i < 18; i++){Bullet.addNormal(getX(), getY(), i * 20 ,6);}
				}
			}
		}else{
			// 打ち込み返し
			if(level > 4 && id == 0){
				if(param > 10){
					if(level < 10){
						for(int i=0;i<3;i++) Bullet.addNormal(getX() + rand()%32 - 16 , getY() + rand()%32 - 16, getAim() + rand()%15 - 7.5, 5 + cast(float)(rand%100)/100);
					}else if(level < 15){
						for(int i=0;i<3;i++) Bullet.addNormal(getX() + rand()%32 - 16 , getY() + rand()%32 - 16, getAim() + rand()%15 - 7.5, 3 + cast(float)(rand%300)/100);
					}else if(level < 20){
						for(int i=0;i<3;i++) Bullet.addNormal(getX() + rand()%32 - 16 , getY() + rand()%32 - 16, getAim() + rand()%15 - 7.5, 1 + cast(float)(rand%500)/100);
					}else{
						for(int i=0;i<5;i++) Bullet.addNormal(getX() + rand()%32 - 16 , getY() + rand()%32 - 16, getAim() + rand()%15 - 7.5, 1 + cast(float)(rand%500)/100);
					}
				}else if(level > 24){
					Bullet.addNormal(getX() + rand()%24 - 12, getY() + rand()%24 - 12, getAim() + rand()%20 - 10, 1+(rand%100)/100);
				}
				}else if(level > 9 &&rand()%2 == 0){
					Bullet.addNormal(getX() + rand()%24 - 12, getY() + rand()%24 - 12, getAim(), 1+(rand%50)/100);
			}
		}
	}
	/**
	 * 消滅
	 */
	void vanish()
	{
		exist = false;
		Particle.addDestroy(getX(), getY(), w/4);
		switch(id)
		{
		case 0:
			Hell_playWAV("damage");
			g_gamemain.state = g_gamemain.STATE_GAMECLEAR;
			g_gamemain.timer = 240;
			Bullet.parent.vanish();
			Item.parent.vanish();
			foreach(enemy; parent.getArray())
			{
				if(!enemy.exist) continue;
				enemy.exist = false;
				Particle.addDestroy(enemy.getX(), enemy.getY(), enemy.w/4);
			}
			break;
		case 1,3, 4, 5, 7, 8, 9, 10, 11, 12,13,15 :
			Item.add(getX(), getY(), 270, -2);
			break;
		default:
			break;
		}
		
		// 自爆打ち返し
		if(level > 19 && hp > 0){
			for(int i = 0; i < 18; i++) Bullet.addNormal(getX(), getY(), getAim() + rand()%10 - 5 ,3 + cast(float)(rand()%50)/100);
		}
	}
	
	/**
	 * 更新
	 */
	void update()
	{
		switch(id)
		{
		case 0:
			// ひたすら敵を生成。CN:Big-Mother.
			timer++;
			int t = timer;
			bool enemy_generate = false;
			if(timer > 2400 || parent.getLives() < 2 || level % 5 == 4){
				enemy_generate = true;
				if(timer > 2400) t = (timer - 2400) * 2;
			}
			
			if(timer == 10 && level > 9){
				if(level > 19){
					add(14, getX(), getY(), 32, 32, 270 + 60, 10);
					add(14, getX(), getY(), 32, 32, 270 - 60, 10);
				}else{
					add(12, getX(), getY(), 32, 32, 270, 10);
				}
			}
			if(t == 60) {add(5, getX(), getY(), 32, 32, 270, 5);}
			if(t == 240 && enemy_generate) {add(15, getX(), getY(), 32, 32, 270 + 60, 10);add(15, getX(), getY(), 32, 32, 270 - 60, 10);}
			if(t == 420 && enemy_generate) {
				if(target.x < g_width / 2) {add(1, getX(), getY(), 32, 32, 270 + 90, 12);add(1, getX(), getY(), 32, 32, 270 + 45, 10);
				}else{add(1, getX(), getY(), 32, 32, 270 - 90, 12);add(1, getX(), getY(), 32, 32, 270 - 45, 10);}
			}
			//if(parent.getLives() < 2 && timer > 120) {add(1, getX(), getY(), 24, 24, 150 + rand()%240, 10);}
			//if(timer == 120) {add(12, getX(), getY(), 32, 32, 180 + rand()%180, 10);}
			if(t == 600 && enemy_generate)  {add(12, getX(), getY(), 32, 32, 180 + rand()%180, 10);}// ボム
			if(t == 780 && enemy_generate)  {add(11, getX(), getY(), 32, 32, 180, 10); add(11, getX(), getY(), 32, 32, 0, 10);}
			if(t == 1140 && enemy_generate)  {add(7, getX(), getY(),16,16, 270, 15);}	// 花ビット
			if(t == 1320 && enemy_generate) {add(8, getX(), getY(), 32, 32, 195, 15); add(8, getX(), getY(), 32, 32, 345, 15);}
			if(t == 1500 && enemy_generate) {add(9, getX(), getY(), 32, 32, 180, 10); add(10, getX(), getY(), 32, 32, 360, 10);}
			if(t == 1680 && enemy_generate)  {foreach(i; [-90, 90]) add(13, getX(), getY(), 32, 32, i+270, 10);}
			if(t == 1860 && enemy_generate) {
				if(target.x < g_width / 2) {add(3, getX(), getY(), 32, 32, 180, 10);
				}else{add(4, getX(), getY(), 32, 32, 0, 10);}
			}
			if(t == 2220 && enemy_generate) {add(15, getX(), getY(), 32, 32, 270 + 30, 10);add(15, getX(), getY(), 32, 32, 270 - 30, 10);}
	
			if(t % 180 == 60) {for(int i=0;i<3;i++)add(2, getX(), getY(), 12, 12, 180 + rand()%180, 2);}
			super.move();
			if(timer == 2400) {
				Particle.addMessage(160, 240, "CRAZY MODE");
				}
			if(timer == 2500){
				add(14, getX(), getY(), 32, 32, 270 + 90, 10);
				add(14, getX(), getY(), 32, 32, 270 - 90, 10);
			}
			if(timer == 2800) {
				add(14, getX(), getY(), 32, 32, 270 + 30, 20);
				add(14, getX(), getY(), 32, 32, 270 - 30, 20);
				add(5, getX(), getY(), 32, 32, 270, 3);
			}
			/*
			// 発狂モード。Crazy-Mother.
			if(timer > 2300) {
				foreach(i;[120,360,600,840,960,990,1020]) if(2300 + i == timer) add(14, getX(), getY(), 32, 32, 270 + 60, 10);
				foreach(i;[240,480,720,840,900,930,1020]) if(2300 + i == timer) add(14, getX(), getY(), 32, 32, 270 - 60, 10);
			}
			
			if(timer == 2300) {Particle.addMessage(160+60, 240, "CRAZY MODE");}
			if(timer == 2300) {add(14, getX(), getY(), 32, 32, 270, 10);}
			if(timer == 2400) {add(7, getX(), getY(), 8, 8, 270, 15);}
			if(timer == 2500) {add(1, getX(), getY(), 32, 32, 225, 10); add(1, getX(), getY(), 32, 32, 315, 10);}
 			if(timer == 2700) {add(8, getX(), getY(), 32, 32, 205, 10); add(8, getX(), getY(), 32, 32, 335, 10);}
			if(timer == 2900) {foreach(i; [0, 30, 60]) add(5, getX(), getY(), 32, 32, i+240, 5);}
			if(timer == 3100) {add(3, getX(), getY(), 32, 32, 182, 10); add(4, getX(), getY(), 32, 32, 2, 10);}
			if(timer == 3150) {add(9, getX(), getY(), 32, 32, 225, 10); add(10, getX(), getY(), 32, 32, 315, 10);}
			if(timer == 3150) {add(3, getX(), getY(), 32, 32, 180, 10); add(4, getX(), getY(), 32, 32, 0, 10);}
			if(timer == 3150) {add(7, getX(), getY(), 8, 8, 270, 15);}
			if(timer == 3200) {add(3, getX(), getY(), 32, 32, 184, 10); add(4, getX(), getY(), 32, 32, 4, 10);}
			if(timer == 3300) {Hell_stopBgm(3000);}
			if(timer == 3600) {Particle.addMessage(getX(), getY(), "SELF DESTRUCTION");}
*/
			if(timer == 3600) vanish();
			break;
		case 1:
			// 停止して高速狙い撃ち弾発射。CN:Silent-Killer.
			timer++;
			if(timer < 120) {vx *= 0.9; vy *= 0.9;}
			else if(timer > 510){vanish();}
			if(timer%45 == 0 && timer > 89) {
				float v = 1.5 + (timer / 90);
				Bullet.add(getX(), getY(), getAim(), v);
				Bullet.add(getX(), getY(), getAim(), v * 0.9);
			}
			break;
		case 2:
			// 高速離脱艦。CN:Lightning-Force.
			timer++;
			if(timer < 90) {vx *= 0.98; vy *= 0.98;}
			else if(timer == 90) {
				Vec2D v = new Vec2D(target.x-x, target.y-y); v.normalize();
				if(level%5 == 1 && level > 19){
					ax = v.x; ay = v.y;
				}else{
					ax = -v.x; ay = -v.y;
				};}
			else if(timer == 91) {
				float r = getAim()+rand()%10-5;
				Bullet.add(getX(), getY(), r, 2);
				Bullet.add(getX(), getY(), r, 2.2);
				Bullet.add(getX(), getY(), r+60, 2);
				Bullet.add(getX(), getY(), r+60, 2.2);
				Bullet.add(getX(), getY(), r-60, 2);
				Bullet.add(getX(), getY(), r-60, 2.2);
			}
			ax *= 0.85;
			ay *= 0.85;
			break;
		case 3, 4:
			// 両サイドから挟み撃ち。CN:Twin-Witch.
			timer++;
			if(timer < 150) {vx *= 0.935; vy *= 0.935;}
			else if(timer == 150) {vx = 0;vy = 4;}
			else if(timer%9 == 0) {
				Bullet.add(getX(), getY(), (id-3)*180, 1, -0.1*(id-3.5));Bullet.add(getX(), getY(), (id-3)*180, 0.8 , -0.1*(id-3.5));
				}
			break;
		case 5:
			// 4-5way。CN:Smooth-Criminal
			timer++;
			if(timer < 20) {vx *= 0.89; vy *= 0.89;}
			else if(timer == 80 || timer == 140 || timer == 200) {foreach(i; [-60, -30, 0, 30, 60]) {Bullet.add(getX(), getY(), i + getAim(), 3);Bullet.add(getX(), getY(), i + getAim(), 2.8);}}
			else if(timer == 110 || timer == 170 || timer == 230) {foreach(i; [-45, -15, 15, 45])     {Bullet.add(getX(), getY(), i + getAim(), 3);Bullet.add(getX(), getY(), i + getAim(), 2.8);}}
			else if(timer > 290){vanish();}
			break;
		case 6:
			// 停止してやや遅い狙い撃ち弾発射。により大量生成。CN:Silent-Killer-Children.
			timer++;
			if(timer < 60) {vx *= 0.9; vy *= 0.9;}
			else if(timer > 155){vy += -0.5;} // {vanish();}
			else if(timer%50 == 0) {
				Bullet.add(getX(), getY(), getAim() - 30, 3);
				//Bullet.add(getX(), getY(), getAim(), 3);
				Bullet.add(getX(), getY(), getAim() + 30, 3);
			}
			break;
		case 7:
			// Silent-Killer-Childrenの種。CN:Grim-Reaper.
			timer++;
			if(timer < 60) {vx *= 0.9; vy *= 0.9;}
			else if(timer == 90) {for(int i = 0; i < 16; i++) add(6, getX(), getY(), 16, 16, i*360/16, 8);}
			else if(timer > 90) {vanish();}
			break;
		case 8:
			// ワインダー。CN:Double-Mermaid.
			timer++;
			if(timer < 120) {vx *= 0.9; vy *= 0.9;}
			else if(timer < 400 && timer%2 == 0) {Bullet.add(getX(), getY(), 270+25*sin(deg2rad((timer-120)*2)), 6);}
			else if(timer > 460){vanish();}
			break;
		case 9, 10:
			// 回転砲台。CN:Love-Fool.
			timer++;
			if(timer < 120) {vx *= 0.9; vy *= 0.9;}
			else if(timer < 400 && timer%4 == 0) {
				Bullet.add(getX(), getY(), 2*(id-9.5)*(timer*3)%360, timer / 60);Bullet.add(getX(), getY(), 2*(id-9.5)*(timer*3)%360, timer / 60 * 0.95);
				}
			
			else if(timer == 460){vanish();}
			break;
		case 11:
			// 縦溝。CN:Black Vice(コスト３)
			timer++;
			if(timer < 60) {vx *= 0.935; vy *= 0.935;}
			else if(timer%25 == 0){
				Bullet.add(getX(), getY(), 270 , 4);Bullet.add(getX(), getY(), 270 , 3.8);
				Bullet.add(getX(), getY(), 270 , 1);Bullet.add(getX(), getY(), 270 , 0.9);
				}
			else if(timer >= 290) {vanish();};
			if(timer == 60) {
				if(x > g_width / 2) vx = -1.5;
				if(x < g_width / 2) vx = 1.5;
				vy = 0;
			}
			break;
		case 12:
			// 全方位ばら撒き単発ボム。CN:Ball Lightning(コスト赤赤赤)
			timer++;
			if(timer < 60) {vx *= 0.9; vy *= 0.9;}
			else if(timer == 90){for(int i = 0; i < 36; i++){Bullet.add(getX(), getY(), i * 10 ,4);Bullet.add(getX(), getY(), i * 10+getAim() ,4.4);}}
			else if(timer >= 195) {vanish();}
			break;
		case 13:
			// シンプル・ザ・ファイヴウエ～イ。CN:ぼんだいん
			timer++;
			if(timer < 60) {vx *= 0.89; vy *= 0.89;}else{vx *= 0.89;}
			if(timer%15 == 0 & y < 240) {foreach(i; [-90, -45, 0, 45, 90]) {Bullet.add(getX(), getY(), i + getAim(), 3);Bullet.add(getX(), getY(), i + getAim(), 2.5);Bullet.add(getX(), getY(), i + getAim(), 2);}}
			if(timer > 200){vx *= 0.8 ;vy *= 0.8;}
			if(timer > 300){vanish();}
			break;
		case 14:
			// 全方位密集。CN:Winter Orb(コスト２)
			timer++;
			if(timer < 60) {vx *= 0.89; vy *= 0.89;}
			else if(timer== 120) {
				for(int i=0;i<360;i+=5){
					Bullet.add(getX(), getY(), (i + 2.5) + getAim(), 9);
					Bullet.add(getX(), getY(), i + getAim(), 2);
				}
			}else if(timer > 180)vanish();
			break;
		case 15:
			// 加速自機狙い垂れ流し。CN:Alladin Ring(コスト８)
			timer++;
			if(timer < 32) {vx *= 0.89; vy *= 0.89;}
			else if(timer%4 == 0 && timer%90 > 60 && timer < 240) {Bullet.add(getX(), getY(), getAim(), (cast(float)timer%90 - 30) / 10);}
			else if(timer > 300)vanish();
			else if(timer > 200){vx *= 0.8 ;vy *= 0.8;}
			break;
		}
		super.move();
		if(isOutOfWindow())
		{
			exist = false;
		}
	}
	/**
	 * 描画
	 */
	void draw()
	{
		super.draw(0, 0xff, 0);
		switch(id)
		{
		case 0:
			int w = 300*hp/hpmax;
			Hell_fillRect(10, 10, w, 10, 0xff*(hpmax-hp)/hpmax, 0xff*hp/hpmax, 0x80);
			break;
		default:
			break;
		}
	}
}

/**
 * アイテムトークン
 */
class Item : Token
{
public:
	/**
	 * 親
	 */
	static TokenManager!(Item) parent;
	int id;     // ID
	
	int timer;
public:
	/**
	 * 生成
	 */
	void create(int id, float x, float y, float w, float h, float deg, float speed, float ax=0, float ay=0)
	{
		this.id = id;
		timer = 0;
		super.create(x, y, w*2.5 , h*2.5 , w , h , cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed, ax, ay);
	}
	/**
	 * 生成（外から呼ばれる）
	 */
	static void add(float x, float y, float deg, float speed, float ax=0, float ay=0)
	{
		Item item = parent.getToken();
		if(!item) return;
		item.create(0, x, y, 16, 16, deg, speed, ax, ay);
	}
	/**
	 * 更新
	 */
	void update()
	{
		super.move();
		timer++;
		vx *= 0.98; vy *= 0.98;
		vy += 0.06;
		if(isOutOfWindow())
		{
			exist = false;
		}
	}
	/**
	 * 描画
	 */
	void draw()
	{
//		dump();
//		super.draw(0xff, 0xff, 0);
		for(int t=0;t<4;t++){
			float gx = 6.0 * sin(cast(float)t * 3.141592 / 2.0 + cast(float)timer * 0.1);
			float gy = 6.0 * cos(cast(float)t * 3.141592 / 2.0 + cast(float)timer * 0.1);
			Hell_fillRect(cast(int)(x - gx - 4) , cast(int)(y - gy - 4) , 8, 8, 0xff , 0xff , 0);
		}
	}
}

/**
 * 打ち返した弾
 */
class Bomb : Token
{
public:
	/**
	 * 親
	 */
	static TokenManager!(Bomb) parent;
public:
	/**
	 * 生成
	 */
	void create(float x, float y, float w, float h, float deg, float speed, float ax=0, float ay=0)
	{
		super.create(x, y, w, h, w, h,cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed, ax, ay);
	}
	/**
	 * 生成（外から呼ばれる方）
	 */
	static void add(float x, float y, float deg, float speed, float ax=0, float ay=0)
	{
		Bomb bomb = parent.getToken();
		if(!bomb) return;
		bomb.create(x, y, 16, 16,deg, speed, ax, ay);
	}
	/**
	 * 消滅
	 */
	void vanish()
	{
		exist = false;
		Particle.addBomb(x, y);
	}
	/**
	 * 更新
	 */
	void update()
	{
		super.move();
		if(isOutOfWindow())
		{
			exist = false;
		}
	}
	/**
	 * 描画
	 */
	void draw()
	{
		super.draw(0xff, rand()%255, rand()%255);
	}
}

/**
 * パーティクル
 */
class Particle : Token
{
public:
	/**
	 * 親
	 */
	static TokenManager!(Particle) parent;
	int id;
	int timer;
	char[] msg;
public:
	void create(int id, float x, float y, float w, float h, float deg=0, float speed=0)
	{
		this.id = id;
		switch(id)
		{
		case 0:
		case 2:
		case 5:
			timer = 80+rand()%40;
			break;
		case 1:
		case 3:
		case 6:
			timer = 60;
			break;
		case 4:
			timer = 4;
			break;
		case 7:
			timer = 16;
			break;
		}
		super.create(x, y, w, h, w,h,cos(deg2rad(deg))*speed, -sin(deg2rad(deg))*speed);
	}

	static void add(int id, float x, float y, float w, float h, float deg=0, float speed=0, char[] msg=null)
	{
		Particle particle = parent.getToken();
		if(!particle) return;
		particle.create(id, x, y, w, h, deg, speed);
		particle.msg = msg;
	}
	static void addDestroy(float x, float y, float size)
	{
		for(int i = 0; i < 8; i++)
		{
			add(0, x, y, size, size, rand()%360, 1+rand()%3);
		}
	}
	static void addItem(float x, float y)
	{
		add(1, x-60, y, 0, 0, 90, 1);
	}
	static void addRecovery(float x, float y)
	{
		add(3, x-60, y, 0, 0, 90, 1);
	}
	static void addDamage(float x, float y, int cnt)
	{
		for(int i = 0; i < cnt; i++)
		{
			add(2, x, y, 4, 4, rand()%360, 1+rand()%3);
		}
	}
	static void addWinder(float x, float y)
	{
		add(4, x, y, Player.BOMB_SIZE, Player.BOMB_SIZE);
	}
	static void addBomb(float x, float y)
	{
		for(int i = 0; i < 4; i++)
		{
			add(5, x, y, 4, 4, rand()%360, 1+rand()%3);
		}
	}
	static void addMessage(float x, float y, char[] msg)
	{
		add(6, x-msg.length*12/2, y, 0, 0, 90, 1, msg);
	}
	static void addBullet(float x, float y)
	{
		add(7, x, y, 8, 8, 0, 0);
	}
	void update()
	{
		switch(id)
		{
		case 0:
		case 2:
		case 5:
			timer--;
			if(timer < 1) exist = false;
			vx *= 0.97;
			vy *= 0.97;
			break;
		default:
			timer--;
			if(timer < 1) exist = false;
			break;
		}
		super.move();
	}
	void draw()
	{
		switch(id)
		{
		case 7:
			if(timer/4 < 1) return;
			Hell_fillRect(cast(int)(x+w/2), cast(int)(y+h/2), timer/4, timer/4 , 0xff);
			break;
		case 0:
			if(timer < 30 && timer%4 != 0) return;
			super.draw(rand()%255, 0xff, rand()%255);
			break;
		case 2:
			if(timer < 30 && timer%4 != 0) return;
			super.draw(rand()%255, rand()%255, 0xff);
			break;
		case 5:
			if(timer < 30 && timer%4 != 0) return;
			super.draw(0xff, rand()%255, rand()%255);
			break;
		case 1:
			int alpha = timer > 30 ? 255 : 255*timer/30;
			Hell_drawFont("POWER UP!", cast(int)x, cast(int)y, alpha);
			break;
		case 3:
			int alpha = timer > 30 ? 255 : 255*timer/30;
			Hell_drawFont("RECOVERY!", cast(int)x, cast(int)y, alpha);
			break;
		case 4:
			super.draw(rand()%255, rand()%255, 0xff);
			break;
		case 6:
			int alpha = timer > 30 ? 255 : 255*timer/30;
			Hell_drawFont(msg, cast(int)x, cast(int)y, alpha);
			break;
		}
	}
}

/**
 * メインシーン
 */
class GameMain
{
public:
	enum
	{
		STATE_INIT,
		STATE_MAIN,
		STATE_DAMAGE_WAIT,
		STATE_POWERUP_WAIT,
		STATE_BOMB_WAIT,
		STATE_GAMEOVER,
		STATE_GAMECLEAR,
		STATE_GAMECLEAR_WAIT,
	}
	int score;
	int state;
	int timer;
	int counter;
	int bonus_ratio;
	int max_bonus_ratio;
	int item_gain;
	
	bool end;
	Player player;
	int tPlaying;
public:
	this()
	{
		player          = new Player();
		Enemy.target    = player;
		Bullet.target   = player;
		Bullet.parent   = new TokenManager!(Bullet)(1024);
		Shot.parent     = new TokenManager!(Shot)(64);
		Enemy.parent    = new TokenManager!(Enemy)(64);
		Particle.parent = new TokenManager!(Particle)(256);
		Item.parent     = new TokenManager!(Item)(64);
		Bomb.parent     = new TokenManager!(Bomb)(64);
		reset();
		//level = 0;
	}
	void reset()
	{
		tPlaying = 0;
		score = 0;
		counter = 0;
		bonus_ratio = 0;
		max_bonus_ratio = 0;
		item_gain = 0;
		state = STATE_INIT;
		Bullet.parent.vanish();
		Shot.parent.vanish();
		Enemy.parent.vanish();
		Particle.parent.vanish();
		Item.parent.vanish();
		Bomb.parent.vanish();
		player.create(g_width/2, g_height*8/10,16 , 16 );
		Enemy.add(0, 160, 30 + 80/2, 80, 80);
		timer = 0;
		end = false;
		Hell_playBgm("resource/sound/bgm.ogg", 3000);
	}
	void update()
	{
		// インチキブラー
//		Hell_fillRect(rand()%2*g_width/2, rand()%2*g_height/2, g_width/2, g_height/2, 0x20, 0x20, 0x20);
		Hell_fillRect(0, 0, 0, 0, 0x20, 0x20, 0x20);

		switch(state)
		{
		case STATE_INIT:
			state = STATE_MAIN;
			break;
		case STATE_MAIN:
			tPlaying++;
			updateMain();
			break;
		case STATE_DAMAGE_WAIT:
			int c = 255 * timer / 60;
			Hell_fillRect(0, 0, 0, 0, c, c, c);
			Particle.parent.update();
			timer--;
			if(timer < 1)
			{
				if(player.exist)
				{
					state = STATE_MAIN;
					foreach(bullet; Bullet.parent.getArray())
					{
						if(!bullet.exist) continue;
						Particle.addBullet(bullet.x, bullet.y);
						bullet.exist = false;
					}
				}
				else
				{
					state = STATE_GAMEOVER;
					Hell_stopBgm(3000);
				}
			}
			break;
		case STATE_POWERUP_WAIT:
		case STATE_BOMB_WAIT:
			Particle.parent.update();
			timer--;
			if(timer < 1) state = STATE_MAIN;
			break;
		case STATE_GAMEOVER:
			Particle.parent.update();
			if(timer > 1) timer--;
			else if(isPushEnter())
			{
				
				reset();
			}
			break;
		case STATE_GAMECLEAR:
			player.update();
			Shot.parent.update();
			Bomb.parent.update();
			Particle.parent.update();
			if(timer > 0)
			{
				float t = (240 - timer)/240.0;
				Hell_fillRect(cast(int)(160-160*t), 0, cast(int)(1+320*t), 480, 0xff, 0xff, 0xff);
				timer = timer*98/100;
			}
			if(timer < 1)
			{
				state = STATE_GAMECLEAR_WAIT;
				timer = 120;
			}
			break;
		case STATE_GAMECLEAR_WAIT:
			if(isPushEnter()) end = true;
			else
			{
				if(timer > 1)
				{
					timer--;
					int c = 0xff*(timer)/120;
					Hell_fillRect(0,0,0,0,c,c,c);
				}
				else
				{
					Hell_fillRect(0,0,0,0,rand()%40,rand()%40,rand()%40);
				}
			}
			break;
		}
		
		draw();
	}
	void addScore(int param)
	{
		score += param;
	}
	int getBonusRatio(){
		return cast(int)(bonus_ratio / 20) + item_gain;
	}
	void updateMain()
	{
		Enemy.parent.update();
		Shot.parent.update();
		Bullet.parent.update();
		Item.parent.update();
		Bomb.parent.update();
		bool destroy = false;
		foreach(enemy; Enemy.parent.getArray())
		{
			if(!enemy.exist) continue;
			foreach(shot; Shot.parent.getArray())
			{
				if(!shot.exist || !enemy.exist) continue;
				if(shot.isHit(enemy))
				{
					enemy.damage(1);
					shot.exist  = false;
					if(!enemy.exist)
					{
						destroy = true;
					//	addScore(enemy.hpmax);
					}
				}
			}
			foreach(bomb; Bomb.parent.getArray())
			{
				if(!bomb.exist || !enemy.exist) continue;
				if(bomb.isHit(enemy))
				{
					state = STATE_BOMB_WAIT;
					timer = 5;
					enemy.damage(40);
					destroy = true;
					bomb.vanish();
					if(!enemy.exist)
					{
					//	addScore(enemy.hpmax);
					}
				}
			}
		}
		if(player.exist)
		{
			foreach(item; Item.parent.getArray())
			{
				if(!item.exist) continue;
				if(item.isHit(player))
				{
					item.exist = false;
					player.powerup();
					state = STATE_POWERUP_WAIT;
					timer = 10;
				}
			}
			foreach(bullet; Bullet.parent.getArray())
			{
				if(!bullet.exist) continue;
				if(bullet.isHit(player))
				{
					player.damage();
					state = STATE_DAMAGE_WAIT;
					timer = 60;
					break;
				}
			}
			if(state != STATE_DAMAGE_WAIT)
			{
				foreach(enemy; Enemy.parent.getArray())
				{
					if(!enemy.exist) continue;
					if(enemy.isHit(player))
					{
						enemy.damage(10);
						player.damage();
						state = STATE_DAMAGE_WAIT;
						timer = 60;
						break;
					}
				}
			}
			player.update();
		}

		if(destroy) Hell_playWAV("destroy1");

		Particle.parent.update();
	}
	void draw()
	{
		if(player.exist)
		{
			player.draw();
		}
		Enemy.parent.draw();
		Shot.parent.draw();
		Bullet.parent.draw();
		Item.parent.draw();
		Bomb.parent.draw();
		Particle.parent.draw();
		Hell_drawFont("SCORE : " ~ std.string.toString(score), 10, 20);
		Hell_drawFont("TIME  : " ~ std.string.toString(tPlaying), 10, 50);
		
		if(max_bonus_ratio < getBonusRatio()) max_bonus_ratio = getBonusRatio();
		Hell_drawFont("RATIO : x" ~ std.string.toString(cast(int)(bonus_ratio / 20)) ~ " + " ~ std.string.toString(item_gain), 10,g_height - 24);
		Hell_drawFont("MAX : x" ~ std.string.toString(max_bonus_ratio), g_width - 12 * 10,g_height - 24);
		
		
		switch(state)
		{
		case STATE_INIT:
			break;
		case STATE_MAIN:
			break;
		case STATE_GAMEOVER:
			Hell_drawFont("GAME OVER", 160-40, 240);
			break;
		case STATE_GAMECLEAR_WAIT:
			Hell_drawFont("GAME CLEAR", 160-40, 240);
			break;
		default:
			break;
		}
		
		// fps 計測
		//Hell_drawFont("FPS" ~ std.string.toString(Hell_getFPS()) , 0,0);
		Hell_drawFont("LEVEL " ~ std.string.toString(level) , 0,0);
	}
}

int g_score;

void gametitle()
{
	Hell_fillRect(0, 0, 0, 0, 0x20, 0x20, 0x20);
	Hell_draw("title",(g_width - 260) / 2, 60);
	Hell_drawFont("PUSH 'Z' or Joystick'1'", 40, g_height - 120);
	Hell_drawFont("LEVEL : " ~std.string.toString(level), 40, g_height - 120 + 24);
	
	static int lv_timer = 0;
	if(lv_timer < 1){
		if(isPressUp())    {level++;lv_timer = 10;}
		if(isPressDown())  {level--;lv_timer = 10;}
		if(isPressRight()) {level+=5;lv_timer = 10;}
		if(isPressLeft())  {level-=5;lv_timer = 10;}
		
		if(level < 0)level = 0;
		if(level > 50)level = 50;
	}else{
		lv_timer--;
	}
}

/**
 * エンドロール
 */
class GameEndRoll
{
public:
	int timer = 0;
	bool end;
	int ptr;
	char[][int] msgArray;
	static const int MESSAGE_MAX = 17;
	char[][] Message = [
		"RECT WINDER NT",
		"",
		"",
		"",
		"Development",
		"",
		"  KENMO HORORO",
		"",
		"NT Arrangement",
		"  OMEGA",
		"",
		"",
		"",
		"Development History",
		"",
		"  2006/11/05 release.",
		"  2006/11/19 NT release.",
		"",
		"",
		"",
		"Development Tool",
		"",
		"  Program Language",
		"    'D'",
		"",
		"  Program Library",
		"    dHell",
		"",
		"",
		"  Text Editor",
		"    Hidemaru",,
		"    TeraPad",
		"",
		"",
//		"  Paint & Photo Retouching",
//		"    Photoshop Elements",
//		"    PictBear",
//		"    EDGE",
//		"",
		"  Font",
		"    Tare Font",
//		"    Mika-Chan Font",
		"",
		"",
		"  Sequence Soft",
		"    Music Creator",
		"",
		"",
		"  Mastering Edition",
		"    SoundEngine",
		"",
		"",
		"",
		"",
		"Special Thanks",
		"  Game Hell 2000",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
	];
public:
	this()
	{
		end = false;
		timer = 0;
		for(int i = 0; i < MESSAGE_MAX; i++)
		{
			msgArray[i] = "";
		}
		msgArray[MESSAGE_MAX-1] = Message[ptr];
		ptr++;
		Hell_playBgm("resource/sound/bgm.ogg");
	}
	void update()
	{
		if(end)
		{
			Hell_fillRect();
			Hell_drawFont("END", 160, 180);
			Hell_drawFont("Your Score :" ~ std.string.toString(g_gamemain.score), 160-70, 240);
			Hell_drawFont("Your Time  :" ~ std.string.toString(g_gamemain.tPlaying), 160-70, 280);
			return;
		}
		if(isPushCancel())
		{
			ptr = Message.length - 1;
		}
		timer += 1;
		if(timer > 41)
		{
			timer = 0;
			for(int i = 1; i < MESSAGE_MAX; i++)
			{
				msgArray[i-1] = msgArray[i];
			}
			ptr++;
			if(ptr >= Message.length)
			{
				Hell_stopBgm(3000);
				end = true;
				return;
			}
			msgArray[MESSAGE_MAX-1] = Message[ptr];
		}
		draw();
	}
	void draw()
	{
		Hell_fillRect();
		for(int i = 0; i < MESSAGE_MAX; i++)
		{
			Hell_drawFont(msgArray[i], 30, -timer + 10 + i*41);
		}
	}
}
enum
{
	STATE_GAME_TITLE,
	STATE_GAME_MAIN,
	STATE_GAME_END,
}

int g_state = STATE_GAME_TITLE;

GameMain g_gamemain;
GameEndRoll g_gameendroll;

void loop()
{
	switch(g_state)
	{
	case STATE_GAME_TITLE:
		gametitle();
		if(isPushEnter())
		{
			if(!g_gamemain)
			{
				g_gamemain = new GameMain();
			}
			else
			{
				g_gamemain.reset();
			}
			g_state = STATE_GAME_MAIN;
		}
		break;
	case STATE_GAME_MAIN:
		g_gamemain.update();
		if(g_gamemain.end)
		{
			g_gameendroll = new GameEndRoll();
			g_state = STATE_GAME_END;
		}
		break;
	case STATE_GAME_END:
		g_gameendroll.update();
		if(g_gameendroll.end)
		{
			if(isPushEnter())
			{
				g_state = STATE_GAME_TITLE;
			}
		}
		break;
	}
}

bool isPressEnter()
{
	if(Hell_isPressKey(HELL_z)) return true;
	if(Hell_isPressJButton(0))  return true;
	return false;
}
bool isPushEnter()
{
	if(Hell_isPushKey(HELL_z)) return true;
	if(Hell_isPushJButton(0))  return true;
	return false;
}
bool isPushCancel()
{
	if(Hell_isPushKey(HELL_x)) return true;
	if(Hell_isPushJButton(1))  return true;
	return false;
}
bool isPressLeft()
{
	if(Hell_isPressKey(HELL_LEFT))    return true;
	if(Hell_isPressJKey(HELL_J_LEFT)) return true;
	return false;
}
bool isPressUp()
{
	if(Hell_isPressKey(HELL_UP))    return true;
	if(Hell_isPressJKey(HELL_J_UP)) return true;
	return false;
}
bool isPressRight()
{
	if(Hell_isPressKey(HELL_RIGHT))    return true;
	if(Hell_isPressJKey(HELL_J_RIGHT)) return true;
	return false;
}
bool isPressDown()
{
	if(Hell_isPressKey(HELL_DOWN))    return true;
	if(Hell_isPressJKey(HELL_J_DOWN)) return true;
	return false;
}

/**
 * リソース読込
 */
void load()
{
	Hell_loadWAV("winder", "resource/sound/winder.wav");
	Hell_loadWAV("damage", "resource/sound/damage.wav");
	Hell_loadWAV("destroy1", "resource/sound/destroy1.wav");
	Hell_loadWAV("kin", "resource/sound/kin.wav");
	
	Hell_loadBMP("title","resource/image/rectwinder.bmp");
}

int Hell_main(char[][] args)
{
	try
	{
		if(args.length == 1)
		{
			Hell_init("/+++ RECT WINDER NT +++/", g_width, g_height);
		}
		else
		{
			Hell_init("/+++ RECT WINDER NT +++/", 640, g_height);
		}
		load();
		while(true)
		{
			loop();
			Hell_update();
			Hell_wait(16); // 60fps
		}
	}
	catch(Exception e)
	{
		Hell_write(e);
	}
	finally
	{
		Hell_quit();
	}
	
	return 0;
}



// コンソールを消すおまじない
version (Win32_release)
{
	private import std.c.windows.windows;
	private import std.string;

	extern (C) void gc_init();
	extern (C) void gc_term();
	extern (C) void _minit();
	extern (C) void _moduleCtor();

	extern (Windows)
	public int WinMain(HINSTANCE hInstance,
		HINSTANCE hPrevInstance,
		LPSTR lpCmdLine,
		int nCmdShow)
	{
		int result;
		gc_init();
		_minit();
		try
		{
			_moduleCtor();
			char exe[4096];
			GetModuleFileNameA(null, exe, 4096);
			char[][1] prog;
			prog[0] = std.string.toString(exe);
			result  = Hell_main(prog ~ std.string.split(std.string.toString(lpCmdLine)));
		}
		catch(Object o)
		{
			printf(o.toString());
			result = 1;
		}
		gc_term();
		return result;
	}
}
else
{
	public int main(char[][] args)
	{
		return Hell_main(args);
	}
}

